UNIT hypergeo;
(*  19 feb 1997    transformations for hypergeometric functions
*)

{---- Interface ---------------------------}

{$D+}

INTERFACE

CONST  Precision = 1E-12;

PROCEDURE Transformation_15_1_2 (var a,b,c,x,m:Double);
PROCEDURE Transformation_15_3_4 (var a,b,c,x,m:Double);
PROCEDURE Transformation_15_3_5 (var a,b,c,x,m:Double);
FUNCTION IfNegInt(x: Double):Boolean;
FUNCTION IfZero(x: Double):Boolean;


{---- Implementation ----------------------}

IMPLEMENTATION

FUNCTION IfZero(x: Double):Boolean;
  BEGIN { FUNCTION IfZero }
    If (Abs(x))<Precision Then IfZero:= True
                          Else IfZero:= False;
  END;  { FUNCTION IfZero }

FUNCTION IfNegInt(x: Double):Boolean;
  BEGIN { FUNCTION IfNegativeInteger }
    If (x<0) AND (Abs(x-Round(x))<Precision) Then IfNegInt:= True
                                             Else IfNegInt:= False;
  END;  { FUNCTION IfNegativeInteger }


PROCEDURE Transformation_15_1_2 (var a,b,c,x,m:Double);
  { 19 Feb 98 , Abramowitz and Stegun, p 556
    If ( -c,-a in N, a>c )
    or ( -c,-b in N, b>c )
    Then the series expansion does not apply, but
        2F1(a,b;c;x)/Gamma(c)
              = (a)_[m+1](b)_[m+1]/(m+1)!x^[m+1] 2F1(a+m+1,b+m+1;m+2;x) }
  VAR  Dum :Double;
       i,j: Word;
  BEGIN { PROCEDURE Transformation_15_1_2 }
    If       (c<Precision) AND (Abs(c-Round(c))<Precision)
    Then If     NOT( (a<0) AND (Abs(a-Round(a))<Precision) AND (a>c) )
            AND NOT( (b<0) AND (Abs(b-Round(b))<Precision) AND (b>c) )
         Then
           Begin
Writeln('15.1.2');
             j:= Abs(Round(c));
             For i:= 0 To j Do
               Begin
                 Dum:= (a+i)*(b+i)/(i+1)*x;
                 m:= m*Dum;
               End;
             a:= a+j+1;
             b:= b+j+1;
             c:= j+2;
             x:= x;
           End;
  END;  { PROCEDURE Transformation_15_1_2 }

PROCEDURE Transformation_15_3_5 (var a,b,c,x,m:Double);
  { 23 Feb 98 , Abramowitz and Stegun, p 559
       2F1:= (1-z)^[-b] 2F1(b,c-a,c,x/(x-1))
  }
  VAR  aa: Double;
  BEGIN { PROCEDURE Transformation_15_3_5 }
    m:= m*Exp(-b*Ln(1-x));
    x:= x/(x-1);
    aa:= b; b:= c-a; a:= aa;
  END;  { PROCEDURE Transformation_15_3_5 }


PROCEDURE Transformation_15_3_4 (var a,b,c,x,m:Double);
  { 19 Feb 98 , Abramowitz and Stegun, p 559
    If ( b<-5,c-b>0 )
    Then numerical problems, use
        2F1(a,b;c;x)
              = (1-x)^[-a] 2F1( a,c-b,c, x/(x-1) )
              = (1-x)^[-a] 2F1( c-b,a,c, x/(x-1) )
    by swapping arguments 1,2 the procedure can be used twice }
  VAR  aa: Double;
  BEGIN { PROCEDURE Transformation_15_3_4 }
    If (b<-5) AND (c-b>0) AND (x<0.5)
    Then BEGIN
Writeln('her');
           m:= m*Exp(-a*Ln(1-x));
           b:=c-b;
           x:= x/(x-1);
         END;
    aa:= a; a:=b; b:= aa;
  END;  { PROCEDURE Transformation_15_3_4 }



END.