(*
         Ŀ
                     This is the source for unit SpecialF            
            This contains various algorithms for special functions   
                                                                     
                                                   
                                 SpecialF.PAS                        
                                 5 April 1999                        
                                                   
                                                                     
           If you modify this program, please                        
                                                                     
              (1) Make sure that you have a copy of the original     
                  file.                                              
                                                                     
              (2) Delete this comment (23 lines).                    
                                                                     
           Original size of file: 23756 bytes.                       
         
             Copyright       Bent Nielsen
                             Nuffield College, Oxford OX1 1NF, UK
                             http://www.nuff.ox.ac.uk/users/nielsen
*)
UNIT SPECIALF;
(*  24 juni 1997    special functions
    update 21 august 1997
    update 5 april 1999
*)

{------------------------------------------}
{---- Interface ---------------------------}
{------------------------------------------}

{$D+}

INTERFACE

  USES HyperGeo;

  CONST  Precision = 1E-9;
  CONST  Euler =0.577215664901532860606512;
         LongIntMax = 2147483647;

  {--- Hyperbolic functions ---}
  FUNCTION  Cosh(x : Double): Double;
  FUNCTION  Sinh(x : Double): Double;
  FUNCTION  SinhDiv(x : Double): Double;

  {---- Gamma Functions -----}
  FUNCTION Binomial(x,n: Byte): LongInt;
  FUNCTION BinomialReal(x: Real; n:Byte): Double;
  FUNCTION Gamma(x: Double) :Double;
  FUNCTION GammaIntegral(lambda, x :Double) :Double;
  FUNCTION DiGamma(x: Double): Double;

  {---- Error Function -----}
  FUNCTION Erf(x: Double) :Double;

  {--- Bessel functions ---}
  FUNCTION I0(x: Double): Double;
  FUNCTION I0large(x: Double): Double;
  FUNCTION I1(x: Double): Double;
  FUNCTION I1large(x: Double): Double;

  {--- Hypergeometric functions ---}
  FUNCTION HypergeometricDivGamma(aa,bb,cc:Real; xx,multiplicator: Double;
                                  var terms, order: Byte) :Double;

  {--- probability functions ---}
  FUNCTION DistChiSquare(DF :Byte; x :Double) :Double;
  FUNCTION DistGamma(s1,s2,x: Double; Control: Byte) :Double;
    { Control 1 : s1,s2 = shape , scale
              2 : s1,s2 = expectation , 2moment
              3 : s1,s2 = expectation , variance
    }
  FUNCTION DistNormal(x: Double): Double;
  FUNCTION QuantileGamma(s1,s2,p: Double; Control: Byte) :Double;
    { Control 1 : s1,s2 = shape , scale
              2 : s1,s2 = expectation , 2moment
              3 : s1,s2 = expectation , variance
    }

  {--- cumulants ---}
  FUNCTION Variance(m1,m2:Double) :Double;
  FUNCTION Skewness(m1,m2,m3:Double) :Double;
  FUNCTION Kurtosis(m1,m2,m3,m4:Double) :Double;
  PROCEDURE NormalityTest(m1,m2,m3,m4: Double; N:LongInt;
                          var vari,skew,kurt,test,pvalue: Double);
  FUNCTION GammaFitThirdMoment(m1,m2:Double) : Double;
  FUNCTION GammaFitFourthMoment(m1,m2:Double) : Double;

{------------------------------------------}
{---- Implementation ----------------------}
{------------------------------------------}

IMPLEMENTATION

{--- Bessel functions ---}

FUNCTION I0(x: Double): Double;
  { 24-6-97 Abrahamovitz and Stegun p 378
    Error: < 1.6 E-7 for x<3.75
           < 1.9 E-7 for x>3.75 }
  VAR Dum1, Dum2, Dum4, Dum8, Res : Double;
BEGIN { FUNCTION I0 }
  Dum1:= Abs(x)/3.75;
  Dum2:= Sqr(Dum1);
  Dum4:= Sqr(Dum2);
  Dum8:= Sqr(Dum4);
  If Dum1 < 1
  Then Begin
      Res:= 1 + 3.5156229*Dum2 + 3.0899424*Dum4 + 1.2067492*Dum2*Dum4;
      Res:= Res + 0.2659732*Dum8 + 0.0360768*Dum8*Dum2 + 0.0045813*Dum8*Dum4;
    End
  Else Begin
      Res:= 0.39894228 + 0.01328592/Dum1 + 0.00225319/Dum2;
      Res:= Res - 0.00157565/Dum2/Dum1 + 0.00916281/Dum4;
      Res:= Res - 0.02057706/Dum4/dum1 + 0.02635537/Dum4/Dum2;
      Res:= Res - 0.01647633/Dum4/Dum2/Dum1 + 0.00392377/Dum8;
      Dum1:= Abs(x);
      Res:= Exp(Dum1)/Sqrt(Dum1)*Res;
    End;
  I0:= Res;
END;  { FUNCTION I0 }

FUNCTION I0large(x: Double): Double;
  { 10-2-97 Abrahamovitz and Stegun p 378
    reports squareroot(x)*exp(-x)*I0(x);
    Error: < 1.6 E-7 for x<3.75
           < 1.9 E-7 for x>3.75
    Error: < 1   E-8 for x>20
           < 1   E-9 for x>30 }
  VAR Dum1, Dum2, Dum4, Dum8, Res : Double;
BEGIN { FUNCTION I0large }
  If x<0
  Then Begin
         Writeln('Unit SpecialF, Func I0large: Negative argument');
         HALT;
       End;
  If x < 3.75 Then Res:= Exp(-x)*Sqrt(x)*I0(x);
  If x < 20
  Then Begin
         Dum1:= Abs(x)/3.75;
         Dum2:= Sqr(Dum1);
         Dum4:= Sqr(Dum2);
         Dum8:= Sqr(Dum4);
         Res:= 0.39894228 + 0.01328592/Dum1 + 0.00225319/Dum2;
         Res:= Res - 0.00157565/Dum2/Dum1 + 0.00916281/Dum4;
         Res:= Res - 0.02057706/Dum4/dum1 + 0.02635537/Dum4/Dum2;
         Res:= Res - 0.01647633/Dum4/Dum2/Dum1 + 0.00392377/Dum8;
       End
  Else Begin
         Dum2:= Sqr(x);
         Dum4:= Sqr(Dum2);
         Res:= 1 + 1/8/x + 9/128/dum2;
         Res:= Res + 75/1024/x/dum2;
         Res:= Res + 3675/32768/dum4;
         Res:= Res + 59535/262144/dum4/x;
         Res:= Res/Sqrt(2*Pi);
       End;
  I0large:= Res;
END;  { FUNCTION I0large }


FUNCTION I1(x: Double): Double;
  { 24-6-97 Abrahamovitz and Stegun p 378
    Error: < 8 E-9 for x<3.75
           < 2.2 E-7 for x>3.75 }
  VAR Dum1, Dum2, Dum4, Dum8, Res : Double;
BEGIN { FUNCTION I1 }
  Dum1:= Abs(x)/3.75;
  Dum2:= Sqr(Dum1);
  Dum4:= Sqr(Dum2);
  Dum8:= Sqr(Dum4);
  If Dum1 < 1
  Then Begin
      Res:= 1/2 + 0.87890594*Dum2 + 0.51498869*Dum4 + 0.15084934*Dum2*Dum4;
      Res:= Res + 0.02658733*Dum8 + 0.00301532*Dum8*Dum2;
      Res:= Res + 0.00032411*Dum8*Dum4;
      Res:= x*Res;
    End
  Else Begin
      Res:= 0.39894228 - 0.03988024/Dum1 - 0.00362018/Dum2;
      Res:= Res + 0.00163801/Dum2/Dum1 - 0.01031555/Dum4;
      Res:= Res + 0.02282967/Dum4/dum1 - 0.02895312/Dum4/Dum2;
      Res:= Res + 0.01787654/Dum4/Dum2/Dum1 - 0.00420059/Dum8;
      Dum1:= Abs(x);
      Res:= Exp(Dum1)/Sqrt(Dum1)*Res;
    End;
  I1:= Res;
END;  { FUNCTION I1 }

FUNCTION I1large(x: Double): Double;
  { 10-2-98 Abrahamovitz and Stegun p 378
    reports squareroot(x)*exp(-x)*I0(x);
    Error: < 8 E-9 for x<3.75
           < 2.2 E-7 for x>3.75
           < 5   E-9 for x>20
           < 1   E-9 for x>30 }
  VAR Dum1, Dum2, Dum4, Dum8, Res : Double;
BEGIN { FUNCTION I1large }
  If x<0
  Then Begin
         Writeln('Unit SpecialF, Func I1large: Negative argument');
         HALT;
       End;
  If x < 3.75 Then Res:= Exp(-x)*Sqrt(x)*I1(x);
  If x < 20
  Then Begin
         Dum1:= Abs(x)/3.75;
         Dum2:= Sqr(Dum1);
         Dum4:= Sqr(Dum2);
         Dum8:= Sqr(Dum4);
         Res:= 0.39894228 - 0.03988024/Dum1 - 0.00362018/Dum2;
         Res:= Res + 0.00163801/Dum2/Dum1 - 0.01031555/Dum4;
         Res:= Res + 0.02282967/Dum4/dum1 - 0.02895312/Dum4/Dum2;
         Res:= Res + 0.01787654/Dum4/Dum2/Dum1 - 0.00420059/Dum8;
       End
  Else Begin
         Dum2:= Sqr(x);
         Dum4:= Sqr(Dum2);
         Res:= 1 - 3/8/x - 15/128/dum2;
         Res:= Res - 105/1024/x/dum2;
         Res:= Res - 4725/32768/dum4;
         Res:= Res - 66297/262144/x/dum4;
         Res:= Res - 2585583/4194304/dum2/dum4;
         Res:= Res/Sqrt(2*Pi);
       End;
  I1large:= Res;
END;  { FUNCTION I1large }


{--- Hyperbolske funktioner ---}

FUNCTION Cosh(x : Double): Double;
BEGIN { FUNCTION  Cosh }
  Cosh := ( Exp(x) + Exp(-x) )/2;
END;  { FUNCTION  Cosh }

FUNCTION SINH(x : Double): Double;
BEGIN { FUNCTION  Sinh }
  Sinh := ( Exp(x) - Exp(-x) )/2;
END;  { FUNCTION  Sinh }

FUNCTION  SinhDiv(x : Double): Double;
BEGIN { FUNCTION  SinhDiv }
  If Abs(x)<0.01
  Then
    SinhDiv:= 1 + x*x/6 + x*x*x*x/120
  Else
    SinhDiv:= Sinh(x)/x;
END;  { FUNCTION  SinhDiv }

{---- Gamma Functions -----}

FUNCTION BinomialReal(x: Real; n:Byte): Double;
  { 18-feb-98
    Binomial coefficient (x,n) }
  VAR        Dum : Double;
             i   : Byte;
  BEGIN { FUNCTION Binomial }
    Dum:= 1;
    For i:= 0 To n-1 Do
      Dum:= Dum*(x-i)/(n-i);
    BinomialReal:= Dum;
  END;  { FUNCTION Binomial }

FUNCTION Binomial(x,n : Byte): LongInt;
  { 22-jun-98 }
  VAR         i : Byte; Dum : LongInt;
  BEGIN { FUNCTION Binomial }
    If n=0
    Then Binomial:= 1
    Else If n<0
         Then Binomial:= 0
         Else Binomial:= Round(BinomialReal(x,n));
  END;  { FUNCTION Binomial }

FUNCTION Gamma(x: Double) :Double;
  { 21-8-97
    Stirlings series applied for x>6 otherwise polynomial expansion.
    Then the relative error seems to be < 3e-7.
    Even for x=15 the relative error seems to be ~ 2e-7. }

  FUNCTION GammaStirling(x: Double) :Double;
    { 21-8-1997,
      Stirlings series, only to be applied for large x
      Abramowitz and Stegun, p. 257 }
    VAR   dum1, dum2 : Double;
    BEGIN { FUNCTION GammaStirling }
      Dum2:= Sqr(x);
      Dum1:= 1 + 1/12/x + 1/288/dum2 - 139/51840/x/dum2 - 571/2488320/Sqr(Dum2);
      Dum1:= Dum1*Exp( -x+(x-1/2)*ln(x) )*Sqrt(2*pi);
      GammaStirling:= Dum1;
    END;  { FUNCTION GammaStirling }

  FUNCTION Gamma12(x: Double) : Double;
    { 21-8-1997,
      Polynomial approximation, 6.1.36, Abramowitz and Stegun, p. 257
      error < 3e-7. }
    CONST  b1= -0.577191652;
           b2=  0.988205891;
           b3= -0.897056937;
           b4=  0.918206857;
           b5= -0.756704078;
           b6=  0.482199394;
           b7= -0.193527818;
           b8=  0.035868343;
    VAR   dum, dum1, dum2, dum4 : double;
    BEGIN { FUNCTION Gamma12 }
      dum1:= x-1;
      dum2:= Sqr(dum1);
      dum4:= Sqr(dum2);
      dum:= 1+ b1*dum1 + b2*dum2 + b3*dum1*dum2 + b4*dum4 + b5*dum4*dum1;
      dum:= dum + b6*dum4*dum2 + b7*dum4*dum2*dum1 + b8*Sqr(dum4);
      Gamma12:= dum;
    END;  { FUNCTION Gamma12 }

  BEGIN { FUNCTION Gamma }
    If x > 171 Then Begin Writeln('Argument for Gamma is too big (SpecialF)');
                          HALT; End;
    If x >= 6
    Then Gamma:= GammaStirling(x)
    Else If x>2
         Then Gamma:= (x-1)*Gamma(x-1)
         Else If x<1
              Then Gamma:= Gamma(x+1)/x
              Else Gamma:= Gamma12(x);
  END;  { FUNCTION Gamma }

FUNCTION GammaIntegral(lambda, x :Double) :Double;
  CONST Tolerance= 1e-7;
  VAR   Error : Boolean;
  FUNCTION GammaIntegralExpansion(lambda, x :Double) :Double;
    { 21-8-97
      Expansion of GammaIntegral int_0^x t^lambda-1 exp(-t)
      Abramowitz and Stegun p 262
      Note that the series is alternating.}
    VAR   dum1, dum2, dum3 : double;
          i : byte;
    BEGIN { FUNCTION GammaIntegralExpansion }
      dum1:= 1;
      i:= 0;
      dum3:= 1/lambda;
      Repeat
        Inc(i);
        dum1:= dum1*(-x)/i;
        dum2:= dum1/(lambda+i);
        dum3:= dum3 + dum2;
      Until (Abs(dum2)<Tolerance) or (i=100);
      GammaIntegralExpansion:= Dum3*Exp(lambda*Ln(x));
      If i=100 Then Error:=True;
    END;  { FUNCTION GammaIntegralExpansion }

  FUNCTION GammaIntegralFraction(lambda, x :Double) :Double;
    { 21-8-97
      Continued Fraction Expansion for GammaIntegral
      int_0^x t^lambda-1 exp(-t) dx
      Abramowitz and Stegun p 263
      Implemented in Fortran by
      B.L.Shea (1988): Chi-squared and incomplete gamma integral,
      Applied Statististics, vol 37, p. 466-473.
      See also Schwartz p 120.      }
    VAR  a,b,d,p0, p1, p2, q0, q1, q2, r1, r2, dum : Double;
         c : Byte;
    BEGIN { FUNCTION GammaIntegralFraction }
      a:= 1-lambda;
      b:= a+x+1;
      c:= 0;
      p1:= 1;
      q1:= x;
      p2:= 1+x;
      q2:= b*x;
      r2:= p2/q2;
      Repeat
        p0:= p1; p1:= p2; q0:= q1; q1:= q2; r1:= r2;
        a:= a+1; b:= b+2; c:= c+1; d:= a*c;
        p2:= b*p1 - d*p0;
        q2:= b*q1 - d*q0;
        If Abs(q2) > 0 Then r2:= p2/q2;
      Until ( Abs(r2-r1)< Tolerance ) or ( Abs(r2-r1)< Tolerance*r2 )
                                      or (c=100);
      dum:= Gamma(lambda);
      dum:= dum-exp(-x+lambda*Ln(x))*r2;
      GammaIntegralFraction:= dum;
      If c=100 Then Error:= True;
    END;  { FUNCTION GammaIntegralFraction }

  BEGIN { FUNCTION GammaIntegral }
    Error:= False;
    If x<0 Then Begin Writeln('Negative argument for GammaIntegral (SpecialF)');
                      HALT; End;
    If x<Tolerance
    Then GammaIntegral:= 0
    Else If (x<=1) or (x<0.9*lambda)
         Then GammaIntegral:= GammaIntegralExpansion(lambda, x)
         Else GammaIntegral:= GammaIntegralFraction(lambda, x);
    If Error Then Begin Writeln('Long loop in Gammaintegral (SpecialF)');
                        HALT; End;
  END;  { FUNCTION GammaIntegral }

FUNCTION DiGamma(x: Double): Double;
  { 23 Feb 98
    Works for x>6 and for x natural number }
  VAR RoundX: Integer;
  FUNCTION DiGammaAsymp(x: Double): Double;
    { 23 Feb 98, Abramowitz and Stegun, p259, 6.3.18
      Error: <E-10 for n=9, <3E-10 for n=8, <1E-9 for n=7, <2E-9 for n=6 }
    VAR Dum,Dum2,Dum4: Double;
    BEGIN { FUNCTION DiGammaAsymp }
      Dum2:= Sqr(x);
      Dum4:= Sqr(Dum2);
      Dum:= Ln(x)-1/2/x-1/12/Dum2;
      DiGammaAsymp:= Dum+(1/120-1/252/Dum2)/Dum4;
    END;  { FUNCTION DiGammaAsymp }
  FUNCTION DiGammaInteger(x:Byte): Double;
  VAR j: Byte; Dum: Double;
    BEGIN { FUNCTION DiGammaInteger }
      Dum:= -Euler;
      For j:= 1 To Abs(x-1) Do Dum:= Dum+1/j;
      DiGammaInteger:= Dum;
    END;  { FUNCTION DiGammaInteger }
  BEGIN { FUNCTION DiGamma }
    If (Abs(x-Round(x))<Precision) AND (x<10)
    Then Begin
           RoundX:= Round(x);
           DiGamma:= DiGammaInteger(RoundX);
         End
    Else If (x>6) Then DiGamma:= DiGammaAsymp(x)
         Else Begin Writeln('Argument for DiGamma = ',x);
                    HALT; End;
       END;  { FUNCTION DiGamma }


{---- Error Funktion -----}
FUNCTION Erf(x: Double) :Double;
  { 21-8-97
    Rational approximation, Abromowitz and Stegun, p 299
    abs error < 1.5e-7 }
  CONST  Tolerance = 1e-7;
         p = 0.3275911;
         a1= 0.254829592;
         a2=-0.284496736;
         a3= 1.421413741;
         a4=-1.453152027;
         a5= 1.061405429;
  VAR    dum,t1,t2,t4 : double;
  BEGIN { FUNCTION Erf }
    If x<0 Then Begin Writeln('Negative argument for Erf (specialf.pas)');
                      HALT; End;
    If x<Tolerance
    Then Erf:=0
    Else Begin
           t1:= 1/(1+p*x);
           t2:= Sqr(t1);
           t4:= Sqr(t2);
           dum:= a1*t1 + a2*t2 + a3*t1*t2 + a4*t4 + a5*t1*t4;
           erf:= 1-dum*exp(-Sqr(x));
         End;
  END;  { FUNCTION Erf }

{--- Hypergeometriske funktioner ---}
FUNCTION HypergeometricDivGamma(aa,bb,cc:Real; xx, multiplicator: Double;
                                var terms, order: Byte) :Double;
  { 18 feb 98
    see Abramowitz and Stegun, p. 556.
    computes multiplicator*2F1(a,b;c;x)/Gamma(c)
    Testet with (-2,-2,-1) , (1,1,2) , (0.5,1,1.5)
    }
  Var a,b,c,x,m,dum1,dum2: Double;
      LocalOrder: Byte;
  FUNCTION Hypergeometric(a,b,c,x,multiplier: Double):Double;
  { 19 Feb 98
    This is for cases with "nice" a,b,c
    "terms" = No Terms in series expansion
    "order" = 10potens of maxterm relative to final expression }
    VAR  Sum, Dum1, Dum2, MaxValue :Double;
         i,j: Word;
         MaxTerms: Integer;
  BEGIN { FUNCTION Hypergeometric }
      Dum1:= m;
      Sum := Dum1;
      MaxValue:= Abs(Dum1);
      MaxTerms:= -1;
      If IfNegInt(a) Then MaxTerms:= -Trunc(a);
      If IfNegInt(b) AND (b>a) Then MaxTerms:= -Trunc(b);
      i:= 0;
                  { Writeln('i=',i-1:3,' term=',Dum1:22:8,' sum=',Sum:22:8); {}
      If MaxTerms<>0 Then
        Repeat
          i:= i+1;
          Dum2:= (1+(a-1)/i)*(1+(b-1)/i)/(1+(c-1)/i)*x;
          Dum1:= Dum1*Dum2;
          If Abs(Dum1)>MaxValue Then MaxValue:= Dum1;
          Sum := Sum+Dum1;
                  { Writeln('i=',i-1:3,' term=',Dum1:22:8,' sum=',Sum:22:8); {}
        Until ( Abs(Dum1)< Precision ) OR (i=MaxTerms);
      Hypergeometric:=Sum;

      Terms:= i;
      Order:=0;
      If (MaxValue/10)>Abs(Sum)
      Then Begin Order:=1;
                 If (MaxValue/100)>Abs(Sum)
                 Then Begin Order:=2;
                            If (MaxValue/1000)>Abs(Sum)
                            Then Begin Order:=3;
                                       If (MaxValue/10000)>Abs(Sum)
                                       Then Order:=4;
                                 End;
                      End;
           End;
  END;  { FUNCTION Hypergeometric }

  BEGIN { FUNCTION HypergeometricDivGamma }
    a:= aa; b:= bb; c:= cc; x:= xx; m:= multiplicator;  {local variables}

    If IfZero(x) Then Begin HypergeometricDivGamma:= m/Gamma(c); Exit; End;

    If NOT(IfNegInt(a)) AND IfNegInt(c-a) Then Transformation_15_3_5(a,b,c,x,m);
    If NOT(IfNegInt(a)) AND IfNegInt(c-a) Then Transformation_15_3_5(a,b,c,x,m);

    IF b<a Then Begin Dum1:= b; b:= a; a:= Dum1; End;

    Transformation_15_1_2(a,b,c,x,m);
{    Transformation_15_3_4(a,b,c,x,m);
    Transformation_15_3_4(a,b,c,x,m);           {}

    HypergeometricDivGamma:=Hypergeometric(a,b,c,x,m)/Gamma(c);
  END;  { FUNCTION HypergeometricDivGamma }


{--- sandsynligheds funktioner ---}
FUNCTION DistNormal(x: Double): Double;
  { 21-8-97 Distribution function for standard normal }
  BEGIN { FUNCTION DistNormal }
    If Abs(x)<Precision
    Then DistNormal:= 1/2
    Else If x>0
         Then If x<10
              Then DistNormal:= (1+Erf(x/Sqrt(2)))/2
              Else DistNormal:= 1
         Else DistNormal:= 1-DistNormal(-x);
  END;  { FUNCTION DistNormal }

FUNCTION DistGammaSaddlePoint(shape,x : Double) : Double;
  {
    19.4.1999
    Integral int_x^infty y^(shape-1) exp(-y) dy/Gamma(shape)
    evaluated by saddlepoint approximation
    See J L Jensen: Saddlepoint approximations, OUP 1995, p 72
  }
  VAR    gam,eta,dum,y,x0,x1 : Double;
  BEGIN { FUNCTION DistGammaSaddlePoint }
    gam:= shape-1;
    dum:= x/shape;
    eta:= Sqrt(2*(dum-Ln(dum)-1)); If dum<1 Then eta:= -eta;
    x1:= 0.9;
    Repeat
      x0:= x1;
      dum:= x0 - Ln(x0) -1 - Sqr(eta)/2;
      dum:= dum*x0/(x0-1);
      x1:= x0 - dum;
    Until abs(x0-x1)<precision;
    y:= x1;
    dum:= (eta*y/(y-1) - 1)/eta;
(*    dum:= dum*(1-1/12/shape);
*)    dum:= dum*Exp(-Sqr(eta)/2*gam);
    dum:= dum/Sqrt(2*Pi*gam);
    dum:= 1+dum-DistNormal(eta*Sqrt(gam));
    DistGammaSaddlePoint:= dum
  END;  { FUNCTION DistGammaSaddlePoint }

FUNCTION DistGammaWilsonHilferty(shape,scale,x : Double) : Double;
  { 19.4.1999
    Gamma distribution function.
    See Johnson, Kotz and Balakrishnan: Continuous Univariate Distributions
        vol. 1, 2nd edt, Wiley 1994
  }
  VAR    dum,df,xScaled : Double;
  BEGIN { FUNCTION DistGammaWilsonHilferty }
    df:= 2*shape;
    xScaled:= x*2/Scale;
    dum:= Exp(Ln(xScaled/df)/3) - 1 + 2/9/df;
    dum:= dum*Sqrt(9*df/2);
    DistGammaWilsonHilferty:= DistNormal(dum);
  END;  { FUNCTION DistGammaWilsonHilferty }

FUNCTION DistGamma(s1,s2,x: Double; Control: Byte) :Double;
  { Control 1 : s1,s2 = shape , scale
            2 : s1,s2 = expectation , 2moment
            3 : s1,s2 = expectation , variance
  }
  VAR  shape, scale, dum : Double;
  BEGIN { FUNCTION DistGamma }
    Case Control Of
     1 : Begin Shape:= s1; Scale:= s2 End;
     2 : Begin Scale:= s2/s1-s1; Shape:= s1/scale; End;
     3 : Begin Scale:= s2/s1; Shape:= s1/scale; End;
    End;
    If (Shape > 20) And (x/scale<0.9*shape) And (x>1)
    Then dum:= DistGammaWilsonHilferty(shape,scale,x)
    Else Begin
           dum:= GammaIntegral(shape,x/scale);
           dum:= dum/Gamma(shape);
         End;
    DistGamma:= dum
  END;  { FUNCTION DistGamma }

FUNCTION DistChiSquare(DF :Byte; x :Double) :Double;
  { P(chi<x) }
  BEGIN { FUNCTION DistChiSquare }
    DistChiSquare:= DistGamma(DF/2,2,x,1);
  END;  { FUNCTION DistChiSquare }

FUNCTION QuantileGamma(s1,s2,p: Double; Control: Byte) :Double;
  { Control 1 : s1,s2 = shape , scale
            2 : s1,s2 = expectation , 2moment
            3 : s1,s2 = expectation , variance
  }
  VAR dum1,dum2,x0,x1,shape,scale: Double;
  BEGIN { FUNCTION QuantileGamma }
    Case Control Of
     1 : Begin Shape:= s1; Scale:= s2 End;
     2 : Begin Scale:= s2/s1-s1; Shape:= s1/scale; End;
     3 : Begin Scale:= s2/s1; Shape:= s1/scale; End;
    End;
    Dum1:= Exp(shape*Ln(scale))*Gamma(shape);
    x1:= shape*scale;
    Repeat
      x0:= x1;
      dum2:= Exp((shape-1)*Ln(x0)-x0/scale);
      x1:= x0 - (DistGamma(shape,scale,x0,1)-p)*Dum1/Dum2;
    Until abs(x0-x1)<1E-5;
    QuantileGamma:= x1;
  END;  { FUNCTION QuantileGamma }

  {--- kummulanter ---}

  FUNCTION Variance(m1,m2:Double) :Double;
    { 26jan98 computes variance }
    BEGIN { FUNCTION Variance }
      Variance:= m2- Sqr(m1);
    END;  { FUNCTION Variance }

  FUNCTION Skewness(m1,m2,m3:Double) :Double;
    { 26jan98 computes skewness }
    BEGIN { FUNCTION Skewness }
      Skewness:= m3 - 3*m1*m2 + 2*m1*Sqr(m1);
    END;  { FUNCTION Skewness }

  FUNCTION Kurtosis(m1,m2,m3,m4:Double) :Double;
    { 26jan98 computes kurtosis }
    BEGIN { FUNCTION Kurtosis }
      Kurtosis:= m4 - 4*m1*m3 - 3*Sqr(m2) + 12*Sqr(m1)*m2 - 6*Sqr(Sqr(m1));
    END;  { FUNCTION Kurtosis }

  PROCEDURE NormalityTest(m1,m2,m3,m4: Double; N:LongInt;
                          var vari,skew,kurt,test,pvalue: Double);
    { 25nov98 computes skewness, kurtosis, BowmanShenton test for normality
      Large Sample Test, otherwise see Doornik & Hansen
    }
    VAR  skew1,kurt1,crit,vari1 : Double;
    BEGIN { PROCEDURE NormalityTest }
      vari1:= variance(m1,m2);        vari:=vari1;
      skew1:= Skewness(m1,m2,m3);     skew:=skew1;
      kurt1:= Kurtosis(m1,m2,m3,m4);  kurt:=kurt1;
      crit:= Sqr(skew1)/vari1/Sqrt(vari1)/6+Sqr(kurt1/vari1)/24;
      crit:= N*crit;                 test:=crit;
      pvalue:= 1-DistChiSquare(2,crit);
    END;  { PROCEDURE NormalityTest }


  FUNCTION GammaFitThirdMoment(m1,m2:Double) : Double;
    { 26jan98, Fits Gamma distr. to first two moments, computes third moment }
    Var beta, LambdaPlusOne: double;
    BEGIN
      Beta:= m2/m1-m1;
      LambdaPlusOne:= 1/(1-Sqr(m1)/m2);
      GammaFitThirdMoment:= m2*beta*(LambdaPlusOne+1);
    END;

  FUNCTION GammaFitFourthMoment(m1,m2:Double) : Double;
    { 26jan98, Fits Gamma distr. to first two moments, computes fourth moment }
    Var beta, LambdaPlusOne: double;
    BEGIN
      Beta:= m2/m1-m1;
      LambdaPlusOne:= 1/(1-Sqr(m1)/m2);
      GammaFitFourthMoment:= m2*Sqr(beta)*(LambdaPlusOne+1)*(LambdaPlusOne+2);
    END;

END.